"""
Module realizes plugin manager
"""
import importlib
import itertools
import logging
import sys
from pathlib import Path
from typing import Iterable, NamedTuple, Optional, Type

from blinker import signal
from vlutils.jobs.async_runner import AsyncRunner

from .plugins_meta.base_plugins import BaseBackgroundPlugin, BaseEventPlugin, BasePlugin
from .plugins_meta.plugin_meta import PluginMeta


class ActiveEventPlugin(NamedTuple):
    """
    Container got active event plugin
    """

    # plugin name
    name: str
    # plugin instance
    plugin: BaseEventPlugin


class ActivePluginStore:
    """
    Container for active plugins

    Attributes:
        _plugins (Dict[str, List[BasePlugin]]): a map of plugin module name to a list of plugins which contain
                                                in the module
        _eventNameToPluginsMap (Dict[str, List[ActiveEventPlugin]]): a map of event name and plugins for processing
                                                                     corresponding event
        _backgroundPluginsMap (Dict[str, List[BaseBackgroundPlugin]]): a map of plugin module name to a list of
                                                                       background plugins
    """

    def __init__(self):
        self._plugins: dict[str, list[BasePlugin]] = {}
        self._eventNameToPluginsMap: dict[str, list[ActiveEventPlugin]] = {}
        self._backgroundPluginsMap: dict[str, list[BaseBackgroundPlugin]] = {}
        self.pluginAdded = signal("plugin-added")

    def add(self, pluginModuleName: str, plugins: Iterable[BasePlugin]) -> None:
        """
        Add active plugins to store

        Args:
            pluginModuleName: plugin module name
            plugins: initialized plugins
        """
        self._plugins.setdefault(pluginModuleName, []).extend(plugins)
        for plugin in plugins:
            if isinstance(plugin, BaseEventPlugin):
                self._eventNameToPluginsMap.setdefault(plugin.eventName, []).append(
                    ActiveEventPlugin(name=pluginModuleName, plugin=plugin)
                )
            elif isinstance(plugin, BaseBackgroundPlugin):
                self._backgroundPluginsMap.setdefault(pluginModuleName, []).append(plugin)

            self.pluginAdded.send(plugin)

    def remove(self, pluginModuleName: str) -> None:
        """
        Remove plugins by module name.

        Args:
            pluginModuleName: plugin module name
        """
        # remove from event plugins
        for eventName, plugins in self._eventNameToPluginsMap.items():
            self._eventNameToPluginsMap[eventName] = list(filter(lambda pl: pl.name != pluginModuleName, plugins))
        # remove from background plugins
        self._backgroundPluginsMap.pop(pluginModuleName)
        # remove from all stored plugins
        self._plugins.pop(pluginModuleName)

    def getEventPlugins(self, eventName: str) -> Iterable[BaseEventPlugin]:
        """
        Get active event plugins by event name

        Args:
            eventName: event name

        Returns:
            generator by plugins
        """
        return (activePlugin.plugin for activePlugin in self._eventNameToPluginsMap.get(eventName, []))

    @property
    def backgroundPlugins(self) -> list[BaseBackgroundPlugin]:
        """
        Get list of background plugins

        Returns:
            list of plugins
        """
        return list(itertools.chain(*self._backgroundPluginsMap.values()))

    @property
    def plugins(self) -> list[BasePlugin]:
        """
        Get all active plugins

        Returns:
            all active plugins
        """
        return list(itertools.chain(*self._plugins.values()))

    def getPluginNames(self) -> list[str]:
        """
        Get all plugin module names

        Returns:
            module names
        """
        return list(self._plugins)


class PluginManager:
    """
    Plugin manager

    Attributes:
        pluginsRunner (AsyncRunner): caller plugins
        app (LunaApplication): application
        logger (Logger): plugin logger
        importedPlugins Dict[str, Set[Type[BasePlugin]]]: a map module name (plugin name) to list of loaded plugins
        _pluginsDir (Path): directory with plugins
        _activePluginStore (ActivePluginStore):  active plugin container
    """

    def __init__(self, app: "LunaApplication", pluginsDir: Path):
        sys.path.append(str(pluginsDir))
        self._pluginsDir = pluginsDir
        self.pluginsRunner: Optional[AsyncRunner] = None
        self.logger = logging.getLogger("luna-plugins.plugin_manger")
        self.app = app
        self.importedPlugins: dict[str, set[Type[BasePlugin]]] = {}
        self._activePluginStore = ActivePluginStore()
        self.pluginAdded = self._activePluginStore.pluginAdded

    @property
    def activePlugins(self) -> list[BasePlugin]:
        """
        Active plugins property

        Returns:
            list of active plugins
        """
        return self._activePluginStore.plugins

    def importPlugins(self, pluginModuleName: str) -> set[Type[BasePlugin]]:
        """
        import plugins from modules.

        If module was load already will be return plugins classes from cache.

        Args:
            pluginModuleName: plugin module name

        Returns:
            all plugins from module
        """
        if pluginModuleName not in self.importedPlugins:
            try:
                importlib.import_module(pluginModuleName)
            except Exception as exc:  # pylint: disable-msg=broad-except
                self.logger.error(f"failed to import plugin module '{pluginModuleName}': {exc!r}")
                self.importedPlugins.setdefault(pluginModuleName, set())
            else:
                self.logger.info(f"plugin module '{pluginModuleName}' has been imported")
                importedPlugins = set(
                    plug for plug in PluginMeta.registryPlugins if pluginModuleName in plug.__module__
                )
                self.importedPlugins[pluginModuleName] = importedPlugins
        return self.importedPlugins[pluginModuleName]

    async def initialize(self, pluginsNames: list[str]) -> None:
        """
        Initialize manager (async runner and plugins).

        Args:
            pluginsNames: list of plugin module names
        """
        self.pluginsRunner = AsyncRunner(1000)

        for pluginName in set(pluginsNames) - set(self._activePluginStore.getPluginNames()):
            pluginClasses = self.importPlugins(pluginModuleName=pluginName)
            plugins = [pluginClass(self.app) for pluginClass in pluginClasses]
            initializedPlugins = await self._initializePlugins(plugins)

            self._activePluginStore.add(pluginModuleName=pluginName, plugins=initializedPlugins)

    async def sendEventToPlugins(self, eventName: str, *args, **kwargs) -> None:
        """
        Send event to plugins.

        Args:
            eventName: event name
            *args: positional arguments for event handler function
            **kwargs: named arguments for event handler function
        """
        eventPlugins = self._activePluginStore.getEventPlugins(eventName)
        self.pluginsRunner.runNoWait([self._callEventPlugin(plugin, *args, **kwargs) for plugin in eventPlugins])

    async def _callEventPlugin(self, plugin: BaseEventPlugin, *args, **kwargs) -> None:
        """
        Call event plugin.

        Args:
            plugin: event plugin
        """
        try:
            await plugin.handleEvent(*args, **kwargs)
        except Exception as exc:  # pylint: disable-msg=broad-except
            self.logger.error(
                f"exception in plugin '{plugin.pluginName}', event '{plugin.eventName}', exception: {exc!r}"
            )
        else:
            self.logger.debug(f"success call plugin '{plugin.pluginName}', event '{plugin.eventName}'")

    async def _startBackgroundPlugins(self) -> None:
        """
        Start background plugins.
        """
        for plugin in self._activePluginStore.backgroundPlugins:
            try:
                await plugin.start()
            except Exception as exc:  # pylint: disable-msg=broad-except
                self.logger.error(f"failed to start plugin '{plugin.pluginName}': {exc!r}")

    async def _initializePlugins(self, plugins: list[BasePlugin]) -> list[BasePlugin]:
        """
        Initialize all plugins

        Args:
            plugins: list of plugins to initialize

        Returns:
            list of initialized plugins
        """
        initializedPlugins = []
        for plugin in plugins:
            try:
                await plugin.initialize()
            except Exception as exc:  # pylint: disable-msg=broad-except
                self.logger.error(f"failed to initialize plugin '{plugin.pluginName}': {exc!r}")
            else:
                self.logger.info(f"plugin '{plugin.pluginName}' has been initialized")
                initializedPlugins.append(plugin)

        return initializedPlugins

    async def _stopActivePlugins(self) -> None:
        """
        Stop all active plugins.
        """
        for plugin in self.activePlugins:
            try:
                await plugin.close()
            except Exception as exc:  # pylint: disable-msg=broad-except
                self.logger.error(f"failed to stop plugin '{plugin.pluginName}': {exc!r}")
            else:
                self.logger.info(f"plugin '{plugin.pluginName}' has been stopped")

    async def start(self) -> None:
        """
        Start plugins.
        """
        await self._startBackgroundPlugins()

    async def shutdown(self) -> None:
        """
        Graceful plugins shutdown. Stop all running event plugins coroutines, close stop plugins.
        """
        if self.pluginsRunner is not None:
            await self.pluginsRunner.close()
        await self._stopActivePlugins()
