"""The module implements the service structure."""
from abc import ABC, abstractmethod
from enum import Enum
from typing import ClassVar, Literal, Optional, Union

from attr import define, fields
from stringcase import snakecase
from vlutils.descriptors.data import DescriptorType


@define(slots=True)
class _BaseResult:
    """Base for matching results"""

    _aliases: ClassVar[dict[str, str]]

    def __init_subclass__(cls, **kwargs):
        """Extend subclass with default attribute."""
        super().__init_subclass__(**kwargs)
        cls._aliases = {snakecase(clsField.name): clsField.name for clsField in fields(cls)}

    def asDict(self, targets: set[str]) -> dict:
        """
        Recursively cast class to dict.

        Args:
            targets: list of required fields to return

        Returns:
            self as dict
        """

        res = {}
        for field in targets:
            res[field] = getattr(self, self._aliases[field])
        return res


@define(slots=True)
class Face(_BaseResult):
    """Simple face"""

    #: face id
    faceId: Optional[str] = None
    #: account id
    accountId: Optional[str] = None
    #: avatar
    avatar: Optional[str] = None
    #: linked list ids
    lists: Optional[list[str]] = None
    #: event id
    eventId: Optional[str] = None
    #: user data
    userData: Optional[str] = None
    #: create time
    createTime: Optional[str] = None
    #: external id
    externalId: Optional[str] = None


@define(slots=True)
class FaceMatchResult(_BaseResult):
    """Simple face match result"""

    #: face
    face: Face
    #: similarity
    similarity: Optional[float] = None

    def asDict(self, targets: set[str], withSimilarity=True) -> dict:
        """Represent face match result as dict"""
        if withSimilarity:
            return {"face": self.face.asDict(targets), "similarity": self.similarity}
        return {"face": self.face.asDict(targets)}


@define(slots=True)
class Event(_BaseResult):
    """Simple event"""

    #: create time
    createTime: Optional[str] = None
    #: end time
    endTime: Optional[str] = None
    #: event id
    eventId: Optional[str] = None
    #: handler id
    handlerId: Optional[str] = None
    #: account id
    accountId: Optional[str] = None
    #: external id
    externalId: Optional[str] = None
    #: source
    source: Optional[str] = None
    # stream id
    streamId: Optional[str] = None
    #: top match face/event
    topMatch: Optional[dict] = None
    #: match results
    matchResult: Optional[list[dict]] = None
    #: face detections
    faceDetections: Optional[list] = None
    #: body detections
    bodyDetections: Optional[list] = None
    #: face id
    faceId: Optional[str] = None
    #: attach results
    attachResult: list[str] = None
    #: gender
    gender: Optional[int] = None
    #: age
    age: Optional[int] = None
    #: predominant emotion
    emotion: Optional[int] = None
    #: predominant mask
    mask: Optional[int] = None
    #: reference ethnicity
    ethnicGroup: Optional[int] = None
    #: liveness
    liveness: Optional[int] = None
    #: user data
    userData: Optional[str] = None
    #: location object
    location: Optional[dict] = None
    #: track id
    trackId: Optional[str] = None
    #: tags
    tags: Optional[list[str]] = None
    #: body basic attributes
    bodyBasicAttributes: Optional[dict] = None
    #: body upper attributes
    upperBody: Optional[dict] = None
    #: body lower attributes
    lowerBody: Optional[dict] = None
    #: body accessories
    accessories: Optional[dict] = None
    #: user defined data
    meta: Optional[dict] = None


@define(slots=True)
class EventMatchResult(_BaseResult):
    """Simple event match result"""

    #: event
    event: Event
    #: similarity
    similarity: Optional[float] = None

    def asDict(self, targets: set[str], withSimilarity=True) -> dict:
        """Represent event match result as dict"""
        if withSimilarity:
            return {"event": self.event.asDict(targets), "similarity": self.similarity}
        return {"event": self.event.asDict(targets)}


@define(slots=True)
class Attribute(_BaseResult):
    """Simple attribute"""

    #: create time
    createTime: Optional[str] = None
    #: attribute id
    attributeId: Optional[str] = None
    #: account id
    accountId: Optional[str] = None
    #: basic attributes
    basicAttributes: Optional[dict] = None
    #: basic attributes samples
    basicAttributesSamples: Optional[list[str]] = None
    #: face descriptor samples
    faceDescriptorSamples: Optional[list[str]] = None


@define(slots=True)
class AttributeMatchResult(_BaseResult):
    """Simple attribute match result"""

    #: attribute
    attribute: Attribute
    #: similarity
    similarity: Optional[float] = None

    def asDict(self, targets: set[str], withSimilarity=True) -> dict:
        """Represent attribute match result as dict"""
        if withSimilarity:
            return {"attribute": self.attribute.asDict(targets), "similarity": self.similarity}
        return {"attribute": self.attribute.asDict(targets)}


@define(slots=True)
class ErrorMatchResult:
    """Error match result"""

    #: error code
    errorCode: int
    #: error description
    desc: str
    #: error detail
    detail: str

    @property
    def link(self):
        """Link to the documentation website with the error description."""
        return "https://docs.visionlabs.ai/info/luna/troubleshooting/errors-description/code-{}".format(self.errorCode)

    def asDict(self) -> dict:
        """Represent error match result as dict"""
        return {"error_code": self.errorCode, "desc": self.desc, "detail": self.detail, "link": self.link}


class MatchUnitType(Enum):
    """Candidate and Reference match type enum"""

    #: face
    face = "face"
    #: attribute
    attribute = "attribute"
    #: event
    event = "event"
    #: descriptor
    descriptor = "descriptor"


@define
class Candidate:
    """Candidate class"""

    #: candidate batch filters
    filters: dict
    #: matching targets
    targets: set[str]
    #: matching sort order (usual "create_time_asc", "create_time_desc", "similarity")
    order: str
    #: matching limit
    limit: int
    #: matching threshold
    threshold: float

    @property
    def type(self) -> MatchUnitType:
        """Candidate type property"""
        origin = self.filters["origin"]
        if origin == "faces":
            return MatchUnitType.face
        elif origin == "events":
            return MatchUnitType.event
        elif origin == "attributes":
            return MatchUnitType.attribute

    def asRequestDict(self) -> dict:
        """Represent candidate as dict for request."""
        return {
            "filters": self.filters,
            "targets": list(self.targets),
            "order": self.order,
            "limit": self.limit,
            "threshold": self.threshold,
        }

    def asResponseDict(self) -> dict:
        """Represent candidate as dict for response."""
        return self.filters


@define(slots=True)
class Descriptor:
    """Descriptor class"""

    #: raw descriptor bytes
    descriptor: bytes
    #: descriptor version
    version: int

    def asDict(self) -> dict:
        """Represent descriptor as dict"""
        return {"descriptor": self.descriptor, "version": self.version}


@define(slots=True)
class Reference:
    """Reference class."""

    #: reference id
    id: str
    #: reference type
    type: MatchUnitType
    #: reference descriptor | only if there is raw descriptor specified in request
    descriptor: Optional[Descriptor] = None

    def asRequestDict(self) -> dict:
        """Represent reference as dict for request"""
        if self.descriptor is not None:
            return {"id": self.id, "type": "raw_descriptor", "data": self.descriptor.asDict()}
        return {"id": self.id, "type": self.type.value}

    def asResponseDict(self) -> dict:
        """Represent reference as dict for json response"""
        return {"id": self.id, "type": self.type.value}


@define(slots=True)
class _ObjectIdReference(Reference):
    """
    Reference with object id.

    Serialization:
        associates an external type and its ID with the reference object
    """

    # external reference type
    _externalType: ClassVar[Literal["face_external_id", "event_external_id", "event_track_id"]]
    # primary key for found reference
    _primaryKey: ClassVar[Literal["face_id", "event_id"]]
    # external object id specified in request
    externalObjectId: Optional[str] = None

    def asResponseDict(self) -> dict:
        """Represent reference as dict for json response"""
        if not self.externalObjectId:
            # no reference with given externalObjectId was found, return id as is
            return {"type": self._externalType, "id": self.id}

        # we've successfully found appropriate reference by given object ID
        return {"type": self._externalType, "id": self.externalObjectId, self._primaryKey: self.id}


@define(slots=True)
class FaceExternalIdReference(_ObjectIdReference):
    """Reference with face external id."""

    _externalType = "face_external_id"
    _primaryKey = "face_id"


@define(slots=True)
class EventExternalIdReference(_ObjectIdReference):
    """Reference with event external id."""

    _externalType = "event_external_id"
    _primaryKey = "event_id"


@define(slots=True)
class EventTrackIdReference(_ObjectIdReference):
    """Reference with event track id."""

    _externalType = "event_track_id"
    _primaryKey = "event_id"


@define(slots=True)
class MatchRequest:
    """Matching request container class"""

    #: candidate batch
    candidate: Candidate
    #: reference
    reference: Reference
    #: match descriptor type - face or body
    matchDescriptorType: DescriptorType = DescriptorType.face
    #: account id
    accountId: Optional[str] = None
    #: account id
    requestId: Optional[str] = None

    def asRequestDict(self) -> dict:
        """Represent match request as request to another http-service"""
        return {"candidates": [self.candidate.asRequestDict()], "references": [self.reference.asRequestDict()]}


@define(slots=True)
class MatchResult:
    """Matching result container class."""

    #: match request
    matchRequest: MatchRequest
    #: success match results or error
    results: Union[list[FaceMatchResult], list[EventMatchResult], list[AttributeMatchResult], ErrorMatchResult]

    def asMatches(self) -> dict:
        """Represent matches results as dict"""
        filters = self.matchRequest.candidate.filters
        targets = self.matchRequest.candidate.targets - {"similarity"}
        if isinstance(self.results, ErrorMatchResult):
            return {"filters": filters, "error": self.results.asDict()}
        withSimilarity = "similarity" in self.matchRequest.candidate.targets
        result = [row.asDict(targets=targets, withSimilarity=withSimilarity) for row in self.results]
        return {"filters": filters, "result": result}

    @property
    def successful(self) -> bool:
        """Whether match result is successful"""
        return not isinstance(self.results, ErrorMatchResult)


@define(slots=True)
class RawMatchResult:
    """
    Raw matching result container class. Raw match results are returning `as is` to a client.

    .. warnings:

         If request contains additional targets (which matcher does not supported) service will ignore its.
         The Luna-Python-Matcher-Proxy service does not check a format of returning matches.
    """

    #: match request
    matchRequest: MatchRequest
    #: raw matches from plugin, must be serializable for json and msgpack
    matches: list[dict] = None
    #: match error
    error: dict = None

    def asMatches(self) -> dict:
        """Represent matches results as dict"""
        filters = self.matchRequest.candidate.filters
        if self.error:
            return {"filters": filters, "error": self.error}
        return {"filters": filters, "result": self.matches}

    @property
    def successful(self) -> bool:
        """Whether match result is successful"""
        return not self.error


class IMatcher(ABC):
    """Matcher interface."""

    EMPTY_TARGETS = frozenset([])

    @property
    def availableReferenceTypes(self) -> frozenset[MatchUnitType]:
        """Matcher should implement this."""
        raise NotImplementedError

    @property
    def availableCandidateTypes(self) -> frozenset[MatchUnitType]:
        """Matcher should implement this."""
        raise NotImplementedError

    @property
    def availableDescriptorTypes(self) -> frozenset[DescriptorType]:
        """Matcher should implement this."""
        raise NotImplementedError

    @property
    def availableSortOrder(self) -> frozenset[str]:
        """Matcher should implement this."""
        raise NotImplementedError

    @abstractmethod
    def getAvailableTargets(self, origin: MatchUnitType) -> frozenset[str]:
        """
        Return targets for given origin.

        Helps to distinguish targets like "face_id" - (Event.face_id and Face.face_id).

        Args:
            origin: candidate match type
        """

    @abstractmethod
    def getMatchingCost(self, matchRequest: MatchRequest) -> float:
        """
        Get matching cost.

        Args:
            matchRequest: matching request class

        Returns:
            matching cost
        """

    @abstractmethod
    async def match(self, matchRequest: MatchRequest) -> Union[MatchResult, RawMatchResult]:
        """
        Match method.

        Args:
            matchRequest: matching request class

        Returns:
            matching result
        """
